/**
 * Copyright (c) 2008-2023, MOVES Institute, Naval Postgraduate School (NPS). All rights reserved.
 * This work is provided under a BSD-style open-source license, see project
 * <a href="https://savage.nps.edu/opendis7-java/license.html" target="_blank">license.html</a> and <a href="https://savage.nps.edu/opendis7-java/license.txt" target="_blank">license.txt</a>
 */
// header autogenerated using string template dis7javalicense.txt



package edu.nps.moves.dis7.pdus;

import java.util.*;
import java.io.*;
import edu.nps.moves.dis7.enumerations.*;
import java.nio.ByteBuffer;

/**
 * 5.9.3.1 The IsGroupOf PDU shall communicate information about the individual states of a group of entities, including state information that is necessary for the receiving simulation applications to represent the issuing group of entities in the simulation applications’ own simulation.
 * @see <a href="https://ieeexplore.ieee.org/document/6387564" target="_blank">IEEE Std 1278.1-2012, IEEE Standard for Distributed Interactive Simulation - Application Protocols</a> 
 */
public class IsGroupOfPdu extends EntityManagementFamilyPdu implements Serializable
{

   /** The name of this PDU type */
   public static final String NAME = "IsGroupOfPdu";
   
   /** ID of aggregated entities */
   protected EntityID  groupEntityID = new EntityID(); 

   /** type of entities constituting the group uid 213 */
   protected IsGroupOfGroupedEntityCategory groupedEntityCategory = IsGroupOfGroupedEntityCategory.values()[0];

   /** Number of individual entities constituting the group */
   protected byte numberOfGroupedEntities;

   /** padding */
   protected int pad;

   /** latitude */
   protected double latitude;

   /** longitude */
   protected double longitude;

   /** GED records about each individual entity in the group. Bad specing--the Group Entity Descriptions are not described. */
   protected List< VariableDatum > groupedEntityDescriptions = new ArrayList<>();
 

/** Constructor creates and configures a new instance object */
 public IsGroupOfPdu()
 {
    setPduType( DisPduType.ISGROUPOF );
 }
/** copy method creates a deep copy of current object using preferred marshalling method
 * @return deep copy of PDU */
 public synchronized IsGroupOfPdu copy()
 {
     return copyDataOutputStream();
 }
/** Creates a "deep copy" of current object using ByteBuffer methods.
 * @return deep copy of PDU */
 public synchronized IsGroupOfPdu copyByteBuffer()
 {
     IsGroupOfPdu newCopy = new IsGroupOfPdu();
     ByteBuffer byteBuffer = ByteBuffer.allocate(400);
     try
     {
         this.marshal(byteBuffer);      // working
         newCopy.unmarshal(byteBuffer); // not working
     }
     catch (Exception e)
     {
         System.err.println("IsGroupOfPdu deep copy() marshall/unmarshall ByteBuffer exception " + e.getMessage());
         e.printStackTrace(System.err);
         System.exit(-1); // TODO: Abruptly ending VM not a good practice
     }
     return newCopy;
 }

/** byteArrayOutputStream (baos) is used for marshal/unmarshal serialization
   * @see copyDataOutputStream() */
protected ByteArrayOutputStream byteArrayOutputStream = new ByteArrayOutputStream();
/** dataOutputStream (dos) is used for marshal/unmarshal serialization
   * @see copyDataOutputStream() */
protected DataOutputStream      dataOutputStream      = new DataOutputStream(byteArrayOutputStream);

/** copy method creates a deep copy of current object using DataOutputStream methods.
 * @return deep copy of PDU */
 public synchronized IsGroupOfPdu copyDataOutputStream()
 {
     IsGroupOfPdu newCopy = new IsGroupOfPdu();
     try
     {
         this.marshal(dataOutputStream);
         byte[] byteArrayDOS = byteArrayOutputStream.toByteArray();
         newCopy.unmarshal(ByteBuffer.wrap(byteArrayDOS));
     }
     catch (Exception e)
     {
         System.err.println("IsGroupOfPdu deep copy() marshall/unmarshall DataOutputStream exception " + e.getMessage());
         e.printStackTrace(System.err);
         System.exit(-1); // TODO: Abruptly ending VM not a good practice
     }
        try
        {
                 dataOutputStream.flush();
            byteArrayOutputStream.flush();
            byteArrayOutputStream.reset();
        }
        catch (IOException ioe)
        {
            System.err.println("IsGroupOfPdu copyDataOutputStream() flush IOException: " + ioe.getMessage());
        }
     return newCopy;
 }

  /**
   * Returns size of this serialized (marshalled) object in bytes
   * @see <a href="https://en.wikipedia.org/wiki/Marshalling_(computer_science)" target="_blank">https://en.wikipedia.org/wiki/Marshalling_(computer_science)</a>
   * @return serialized size in bytes
   */
@Override
public synchronized int getMarshalledSize()
{
   int marshalSize = 0; 

   marshalSize = super.getMarshalledSize();
   if (groupEntityID != null)
       marshalSize += groupEntityID.getMarshalledSize();
   if (groupedEntityCategory != null)
       marshalSize += groupedEntityCategory.getMarshalledSize();
   marshalSize += 1;  // numberOfGroupedEntities
   marshalSize += 4;  // pad
   marshalSize += 8;  // latitude
   marshalSize += 8;  // longitude
   if (groupedEntityDescriptions != null)
       for (int idx=0; idx < groupedEntityDescriptions.size(); idx++)
       {
            VariableDatum listElement = groupedEntityDescriptions.get(idx);
            marshalSize += listElement.getMarshalledSize();
       }

   return marshalSize;
}


/** Setter for {@link IsGroupOfPdu#groupEntityID}
  * @param pGroupEntityID new value of interest
  * @return same object to permit progressive setters */
public synchronized IsGroupOfPdu setGroupEntityID(EntityID pGroupEntityID)
{
    groupEntityID = pGroupEntityID;
    return this;
}
/** Getter for {@link IsGroupOfPdu#groupEntityID}
  * @return value of interest */
public EntityID getGroupEntityID()
{
    return groupEntityID;
}


/** Setter for {@link IsGroupOfPdu#groupedEntityCategory}
  * @param pGroupedEntityCategory new value of interest
  * @return same object to permit progressive setters */
public synchronized IsGroupOfPdu setGroupedEntityCategory(IsGroupOfGroupedEntityCategory pGroupedEntityCategory)
{
    groupedEntityCategory = pGroupedEntityCategory;
    return this;
}
/** Getter for {@link IsGroupOfPdu#groupedEntityCategory}
  * @return value of interest */
public IsGroupOfGroupedEntityCategory getGroupedEntityCategory()
{
    return groupedEntityCategory; 
}

/** Setter for {@link IsGroupOfPdu#pad}
  * @param pPad new value of interest
  * @return same object to permit progressive setters */
public synchronized IsGroupOfPdu setPad(int pPad)
{
    pad = pPad;
    return this;
}
/** Getter for {@link IsGroupOfPdu#pad}
  * @return value of interest */
public int getPad()
{
    return pad; 
}

/** Setter for {@link IsGroupOfPdu#latitude}
  * @param pLatitude new value of interest
  * @return same object to permit progressive setters */
public synchronized IsGroupOfPdu setLatitude(double pLatitude)
{
    latitude = pLatitude;
    return this;
}
/** Getter for {@link IsGroupOfPdu#latitude}
  * @return value of interest */
public double getLatitude()
{
    return latitude; 
}

/** Setter for {@link IsGroupOfPdu#longitude}
  * @param pLongitude new value of interest
  * @return same object to permit progressive setters */
public synchronized IsGroupOfPdu setLongitude(double pLongitude)
{
    longitude = pLongitude;
    return this;
}
/** Getter for {@link IsGroupOfPdu#longitude}
  * @return value of interest */
public double getLongitude()
{
    return longitude; 
}

/** Setter for {@link IsGroupOfPdu#groupedEntityDescriptions}
  * @param pGroupedEntityDescriptions new value of interest
  * @return same object to permit progressive setters */
public synchronized IsGroupOfPdu setGroupedEntityDescriptions(List<VariableDatum> pGroupedEntityDescriptions)
{
    groupedEntityDescriptions = pGroupedEntityDescriptions;
    return this;
}
/** Getter for {@link IsGroupOfPdu#groupedEntityDescriptions}
  * @return value of interest */
public List<VariableDatum> getGroupedEntityDescriptions()
{
    return groupedEntityDescriptions; 
}

/**
 * Serializes an object to a DataOutputStream.
 * @throws java.lang.Exception if something goes wrong
 * @see java.io.DataOutputStream
 * @param dos the OutputStream
 */
@Override
public synchronized void marshal(DataOutputStream dos) throws Exception
{
    super.marshal(dos);
    try 
    {
       groupEntityID.marshal(dos);
       groupedEntityCategory.marshal(dos);
       dos.writeByte(groupedEntityDescriptions.size());
       dos.writeInt(pad);
       dos.writeDouble(latitude);
       dos.writeDouble(longitude);

       for (int idx = 0; idx < groupedEntityDescriptions.size(); idx++)
       {
            VariableDatum aVariableDatum = groupedEntityDescriptions.get(idx);
            aVariableDatum.marshal(dos);
       }

    }
    catch(Exception e)
    {
      System.err.println(e);
    }
}

/**
 * Deserializes an object from a DataInputStream.
 * @throws java.lang.Exception if something goes wrong
 * @see java.io.DataInputStream
 * @see <a href="https://en.wikipedia.org/wiki/Marshalling_(computer_science)" target="_blank">https://en.wikipedia.org/wiki/Marshalling_(computer_science)</a>
 * @param dis the InputStream
 * @return marshalled serialized size in bytes
 */
@Override
public synchronized int unmarshal(DataInputStream dis) throws Exception
{
    int uPosition = 0;
    uPosition += super.unmarshal(dis);

    try 
    {
        uPosition += groupEntityID.unmarshal(dis);
        groupedEntityCategory = IsGroupOfGroupedEntityCategory.unmarshalEnum(dis);
        uPosition += groupedEntityCategory.getMarshalledSize();
        numberOfGroupedEntities = (byte)dis.readUnsignedByte();
        uPosition += 1;
        pad = dis.readInt();
        uPosition += 4;
        latitude = dis.readDouble();
        uPosition += 4;
        longitude = dis.readDouble();
        uPosition += 4;
        for (int idx = 0; idx < numberOfGroupedEntities; idx++)
        {
            VariableDatum anX = new VariableDatum();
            uPosition += anX.unmarshal(dis);
            groupedEntityDescriptions.add(anX);
        }

    }
    catch(Exception e)
    { 
      System.err.println(e); 
    }
    return getMarshalledSize();
}

/**
 * Packs an object into the ByteBuffer.
 * @throws java.nio.BufferOverflowException if byteBuffer is too small
 * @throws java.nio.ReadOnlyBufferException if byteBuffer is read only
 * @see java.nio.ByteBuffer
 * @param byteBuffer The ByteBuffer at the position to begin writing
 * @throws Exception ByteBuffer-generated exception
 */
@Override
public synchronized void marshal(java.nio.ByteBuffer byteBuffer) throws Exception
{
   super.marshal(byteBuffer);
   groupEntityID.marshal(byteBuffer);
   groupedEntityCategory.marshal(byteBuffer);
   byteBuffer.put( (byte)groupedEntityDescriptions.size());
   byteBuffer.putInt( (int)pad);
   byteBuffer.putDouble( (double)latitude);
   byteBuffer.putDouble( (double)longitude);

   for (int idx = 0; idx < groupedEntityDescriptions.size(); idx++)
   {
        VariableDatum aVariableDatum = groupedEntityDescriptions.get(idx);
        aVariableDatum.marshal(byteBuffer);
   }

}

/**
 * Unpacks a Pdu from the underlying data.
 * @throws java.nio.BufferUnderflowException if byteBuffer is too small
 * @see java.nio.ByteBuffer
 * @see <a href="https://en.wikipedia.org/wiki/Marshalling_(computer_science)" target="_blank">https://en.wikipedia.org/wiki/Marshalling_(computer_science)</a>
 * @param byteBuffer The ByteBuffer at the position to begin reading
 * @return marshalled serialized size in bytes
 * @throws Exception ByteBuffer-generated exception
 */
@Override
public synchronized int unmarshal(java.nio.ByteBuffer byteBuffer) throws Exception
{
    super.unmarshal(byteBuffer);

    try
    {
        // attribute groupEntityID marked as not serialized
        groupEntityID.unmarshal(byteBuffer);
        // attribute groupedEntityCategory marked as not serialized
        groupedEntityCategory = IsGroupOfGroupedEntityCategory.unmarshalEnum(byteBuffer);
        // attribute numberOfGroupedEntities marked as not serialized
        numberOfGroupedEntities = (byte)(byteBuffer.get() & 0xFF);
        // attribute pad marked as not serialized
        pad = byteBuffer.getInt();
        // attribute latitude marked as not serialized
        latitude = byteBuffer.getDouble();
        // attribute longitude marked as not serialized
        longitude = byteBuffer.getDouble();
        // attribute groupedEntityDescriptions marked as not serialized
        for (int idx = 0; idx < numberOfGroupedEntities; idx++)
        {
        VariableDatum anX = new VariableDatum();
        anX.unmarshal(byteBuffer);
        groupedEntityDescriptions.add(anX);
        }

    }
    catch (java.nio.BufferUnderflowException bue)
    {
        System.err.println("*** buffer underflow error while unmarshalling " + this.getClass().getName());
    }
    return getMarshalledSize();
}

 /*
  * Override of default equals method.  Calls equalsImpl() for content comparison.
  */
@Override
 public synchronized boolean equals(Object obj)
 {
    if(this == obj)
      return true;

    if(obj == null)
       return false;

    if(!getClass().isAssignableFrom(obj.getClass())) //if(getClass() != obj.getClass())
        return false;

    return equalsImpl(obj);
 }

@Override
 public synchronized boolean equalsImpl(Object obj)
 {
     final IsGroupOfPdu rhs = (IsGroupOfPdu)obj;

     if( ! Objects.equals(groupEntityID, rhs.groupEntityID) ) return false;
     if( ! (groupedEntityCategory == rhs.groupedEntityCategory)) return false;
     if( ! (pad == rhs.pad)) return false;
     if( ! (latitude == rhs.latitude)) return false;
     if( ! (longitude == rhs.longitude)) return false;
     if( ! Objects.equals(groupedEntityDescriptions, rhs.groupedEntityDescriptions) ) return false;
    return super.equalsImpl(rhs);
 }

 @Override
 public synchronized String toString()
 {
    StringBuilder sb  = new StringBuilder();
    StringBuilder sb2 = new StringBuilder();
    sb.append(getClass().getSimpleName());
    sb.append(" groupEntityID:").append(groupEntityID); // writeOneToString
    sb.append(" groupedEntityCategory:").append(groupedEntityCategory); // writeOneToString
    sb.append(" pad:").append(pad); // writeOneToString
    sb.append(" latitude:").append(latitude); // writeOneToString
    sb.append(" longitude:").append(longitude); // writeOneToString
    sb.append(" groupedEntityDescriptions: ");
    groupedEntityDescriptions.forEach(r->{ sb2.append(" ").append(r);}); // writeList
    sb.append(sb2.toString().trim());
    // https://stackoverflow.com/questions/2242471/clearing-a-string-buffer-builder-after-loop
    sb2.setLength(0); // reset

   return sb.toString();
 }

 @Override
 public int hashCode()
 {
	 return Objects.hash(this.groupEntityID,
	                     this.groupedEntityCategory,
	                     this.numberOfGroupedEntities,
	                     this.pad,
	                     this.latitude,
	                     this.longitude,
	                     this.groupedEntityDescriptions);
 }
} // end of IsGroupOfPdu
